# Copyright (c) 2023 K.Y. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import queue


class SimpleQueue:
    def __init__(self, maxsize: int):
        self.q = queue.Queue(maxsize=maxsize)

    def put(self, item):
        if not self.q.full():
            self.q.put(item)
        else:
            self.q.get()
            self.q.put(item)

    def get(self, block=True, timeout=None):
        return self.q.get(block=block, timeout=timeout)

    def empty(self):
        return self.q.empty()

    def clear(self):
        while not self.empty():
            self.get()
